""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from onegov.core.upgrade import upgrade_task


@upgrade_task('Rename recommendation columns')
def rename_recommendation_columns(context):
    for old, new in (
        ('ucsp', 'csp'),
        ('sbv', 'sbv_usp'),
        ('cng_travs', 'travs'),
        ('zsa', 'sav')
    ):
        old = f'recommendation_{old}'
        new = f'recommendation_{new}'
        if (
            context.has_column('swissvotes', old) and
            not context.has_column('swissvotes', new)
        ):
            context.operations.alter_column(
                'swissvotes', old, new_column_name=new
            )



from cached_property import cached_property
from collections import OrderedDict
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.file import AssociatedFiles
from onegov.file import File
from onegov.search import ORMSearchable
from onegov.swissvotes import _
from onegov.swissvotes.models.actor import Actor
from onegov.swissvotes.models.localized_file import LocalizedFile
from onegov.swissvotes.models.policy_area import PolicyArea
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import Text
from sqlalchemy.dialects.postgresql import INT4RANGE


LEGAL_FORM = {
    1: _("Mandatory referendum"),
    2: _("Optional referendum"),
    3: _("Popular initiative"),
    4: _("Direct counter-proposal"),
}

RESULT = {
    0: _("Rejected"),
    1: _("Accepted"),
}

RESULT_PEOPLE_ACCEPTED = {
    0: _("Rejected"),
    1: _("Accepted"),
}

RESULT_CANTONS_ACCEPTED = {
    0: _("Rejected"),
    1: _("Accepted"),
    3: _("Majority of the cantons not necessary"),
}

DEPARTMENT_IN_CHARGE = {
    1: _("Federal Department of Foreign Affairs (FDFA)"),
    2: _("Federal Department of Home Affairs (FDHA)"),
    3: _("Federal Department of Justice and Police (FDJP)"),
    4: _("Federal Department of Defence, Civil Protection and Sport (DDPS)"),
    5: _("Federal Department of Finance (FDF)"),
    6: _("Federal Department of Economic Affairs, Education and Research "
         "(EAER)"),
    7: _("Federal Department of the Environment, Transport, Energy and "
         "Communications (DETEC)"),
    8: _("Federal Chancellery (FCh)"),
}

POSITION_FEDERAL_COUNCIL = {
    1: _("Accepting"),
    2: _("Rejecting"),
    3: _("Neutral")
}

POSITION_PARLIAMENT = {
    1: _("Accepting"),
    2: _("Rejecting"),
}

RECOMMENDATION = {
    1: _("Yea"),
    2: _("Nay"),
    3: _("None"),
    4: _("Empty"),
    5: _("Free vote"),
    66: _("Neutral"),
    9999: _("Organization no longer exists"),
}


class SwissVoteFile(File):
    __mapper_args__ = {'polymorphic_identity': 'swissvote'}


class SwissVote(Base, TimestampMixin, AssociatedFiles, ORMSearchable):

    """ A single vote as defined by the code book. """

    __tablename__ = 'swissvotes'

    es_public = True
    es_properties = {
        'title': {'type': 'localized'},
        'keyword': {'type': 'localized'},
        'initiator': {'type': 'localized'},
        'voting_text_de_CH': {'type': 'localized'},
        'voting_text_fr_CH': {'type': 'localized'},
        'federal_council_message_de_CH': {'type': 'localized'},
        'federal_council_message_fr_CH': {'type': 'localized'},
        'parliamentary_debate_de_CH': {'type': 'localized'},
        'parliamentary_debate_fr_CH': {'type': 'localized'},
        # we don't include the voting_booklet, they contain other votes!
    }

    id = Column(Integer, nullable=False, primary_key=True)

    # Formal description
    bfs_number = Column(Numeric(8, 2), nullable=False)
    date = Column(Date, nullable=False)
    decade = Column(INT4RANGE, nullable=False)
    legislation_number = Column(Integer, nullable=False)
    legislation_decade = Column(INT4RANGE, nullable=False)
    title = Column(Text, nullable=False)
    keyword = Column(Text)
    votes_on_same_day = Column(Integer, nullable=False)
    _legal_form = Column('legal_form', Integer, nullable=False)
    initiator = Column(Text)
    anneepolitique = Column(Text)

    # Descriptor
    descriptor_1_level_1 = Column(Numeric(8, 4))
    descriptor_1_level_2 = Column(Numeric(8, 4))
    descriptor_1_level_3 = Column(Numeric(8, 4))
    descriptor_2_level_1 = Column(Numeric(8, 4))
    descriptor_2_level_2 = Column(Numeric(8, 4))
    descriptor_2_level_3 = Column(Numeric(8, 4))
    descriptor_3_level_1 = Column(Numeric(8, 4))
    descriptor_3_level_2 = Column(Numeric(8, 4))
    descriptor_3_level_3 = Column(Numeric(8, 4))

    # Result
    _result = Column('result', Integer)
    result_eligible_voters = Column(Integer)
    result_votes_empty = Column(Integer)
    result_votes_invalid = Column(Integer)
    result_votes_valid = Column(Integer)
    result_votes_total = Column(Integer)
    result_turnout = Column(Numeric(13, 10))
    _result_people_accepted = Column('result_people_accepted', Integer)
    result_people_yeas = Column(Integer)
    result_people_nays = Column(Integer)
    result_people_yeas_p = Column(Numeric(13, 10))
    _result_cantons_accepted = Column('result_cantons_accepted', Integer)
    result_cantons_yeas = Column(Numeric(3, 1))
    result_cantons_nays = Column(Numeric(3, 1))
    result_cantons_yeas_p = Column(Numeric(13, 10))

    # Authorities
    _department_in_charge = Column('department_in_charge', Integer)
    procedure_number = Column(Numeric(8, 3))
    _position_federal_council = Column('position_federal_council', Integer)
    _position_parliament = Column('position_parliament', Integer)
    position_national_council_yeas = Column(Integer)
    position_national_council_nays = Column(Integer)
    position_council_of_states_yeas = Column(Integer)
    position_council_of_states_nays = Column(Integer)

    # Duration
    duration_federal_assembly = Column(Integer)
    duration_post_federal_assembly = Column(Integer)
    duration_initative_collection = Column(Integer)
    duration_initative_federal_council = Column(Integer)
    duration_initative_total = Column(Integer)
    duration_referendum_collection = Column(Integer)
    duration_referendum_total = Column(Integer)
    signatures_valid = Column(Integer)
    signatures_invalid = Column(Integer)

    # Voting recommendations
    _recommendation_fdp = Column('recommendation_fdp', Integer)
    _recommendation_cvp = Column('recommendation_cvp', Integer)
    _recommendation_sps = Column('recommendation_sps', Integer)
    _recommendation_svp = Column('recommendation_svp', Integer)
    _recommendation_lps = Column('recommendation_lps', Integer)
    _recommendation_ldu = Column('recommendation_ldu', Integer)
    _recommendation_evp = Column('recommendation_evp', Integer)
    _recommendation_csp = Column('recommendation_csp', Integer)
    _recommendation_pda = Column('recommendation_pda', Integer)
    _recommendation_poch = Column('recommendation_poch', Integer)
    _recommendation_gps = Column('recommendation_gps', Integer)
    _recommendation_sd = Column('recommendation_sd', Integer)
    _recommendation_rep = Column('recommendation_rep', Integer)
    _recommendation_edu = Column('recommendation_edu', Integer)
    _recommendation_fps = Column('recommendation_fps', Integer)
    _recommendation_lega = Column('recommendation_lega', Integer)
    _recommendation_kvp = Column('recommendation_kvp', Integer)
    _recommendation_glp = Column('recommendation_glp', Integer)
    _recommendation_bdp = Column('recommendation_bdp', Integer)
    _recommendation_mcg = Column('recommendation_mcg', Integer)

    _recommendation_sav = Column('recommendation_sav', Integer)
    _recommendation_eco = Column('recommendation_eco', Integer)
    _recommendation_sgv = Column('recommendation_sgv', Integer)
    _recommendation_sbv_usp = Column('recommendation_sbv_usp', Integer)
    _recommendation_sgb = Column('recommendation_sgb', Integer)
    _recommendation_travs = Column('recommendation_travs', Integer)
    _recommendation_vsa = Column('recommendation_vsa', Integer)

    # Electoral strength
    national_council_election_year = Column(Integer)
    national_council_share_fdp = Column(Numeric(13, 10))
    national_council_share_cvp = Column(Numeric(13, 10))
    national_council_share_sp = Column(Numeric(13, 10))
    national_council_share_svp = Column(Numeric(13, 10))
    national_council_share_lps = Column(Numeric(13, 10))
    national_council_share_ldu = Column(Numeric(13, 10))
    national_council_share_evp = Column(Numeric(13, 10))
    national_council_share_csp = Column(Numeric(13, 10))
    national_council_share_pda = Column(Numeric(13, 10))
    national_council_share_poch = Column(Numeric(13, 10))
    national_council_share_gps = Column(Numeric(13, 10))
    national_council_share_sd = Column(Numeric(13, 10))
    national_council_share_rep = Column(Numeric(13, 10))
    national_council_share_edu = Column(Numeric(13, 10))
    national_council_share_fps = Column(Numeric(13, 10))
    national_council_share_lega = Column(Numeric(13, 10))
    national_council_share_kvp = Column(Numeric(13, 10))
    national_council_share_glp = Column(Numeric(13, 10))
    national_council_share_bdp = Column(Numeric(13, 10))
    national_council_share_mcg = Column(Numeric(13, 10))
    national_council_share_ubrige = Column(Numeric(13, 10))
    national_council_share_yeas = Column(Numeric(13, 10))
    national_council_share_nays = Column(Numeric(13, 10))
    national_council_share_neutral = Column(Numeric(13, 10))
    national_council_share_vague = Column(Numeric(13, 10))

    voting_text = LocalizedFile()
    federal_council_message = LocalizedFile()
    parliamentary_debate = LocalizedFile()
    voting_booklet = LocalizedFile()

    @cached_property
    def legal_form(self):
        return LEGAL_FORM.get(self._legal_form, "")

    @cached_property
    def policy_areas(self):
        def get_level(number, level):
            value = getattr(self, f'descriptor_{number}_level_{level}')
            if value is not None:
                return PolicyArea(value, level)

        result = []
        for number in (1, 2, 3):
            for level in (3, 2, 1):
                area = get_level(number, level)
                if area:
                    result.append(area)
                    break
        return result

    @cached_property
    def department_in_charge(self):
        return DEPARTMENT_IN_CHARGE.get(self._department_in_charge, "")

    @cached_property
    def result(self):
        return RESULT.get(self._result, "")

    @cached_property
    def result_people_accepted(self):
        return RESULT_PEOPLE_ACCEPTED.get(self._result_people_accepted, "")

    @cached_property
    def result_cantons_accepted(self):
        return RESULT_CANTONS_ACCEPTED.get(self._result_cantons_accepted, "")

    @cached_property
    def position_federal_council(self):
        return POSITION_FEDERAL_COUNCIL.get(self._position_federal_council, "")

    @cached_property
    def position_parliament(self):
        return POSITION_PARLIAMENT.get(self._position_parliament, "")

    def position_by_votes(self, attribute):
        yeas = getattr(self, f'{attribute}_yeas')
        nays = getattr(self, f'{attribute}_nays')
        if yeas is None or nays is None:
            return ""
        if yeas == nays:
            return _("Neutral")
        if yeas > nays:
            return _("Accepting")
        return _("Rejecting")

    @cached_property
    def position_national_council(self):
        return self.position_by_votes('position_national_council')

    @cached_property
    def position_council_of_states(self):
        return self.position_by_votes('position_council_of_states')

    def group_recommendations(self, recommendations):
        result = {}
        for actor, recommendation in recommendations:
            if recommendation is not None and recommendation != 9999:
                result.setdefault(recommendation, []).append(actor)

        return OrderedDict([
            (RECOMMENDATION[key], result[key])
            for key in sorted(result.keys())
        ])

    @cached_property
    def recommendations_parties(self):
        return self.group_recommendations((
            (Actor('bdp'), self._recommendation_bdp),
            (Actor('cvp'), self._recommendation_cvp),
            (Actor('edu'), self._recommendation_edu),
            (Actor('evp'), self._recommendation_evp),
            (Actor('fdp'), self._recommendation_fdp),
            (Actor('fps'), self._recommendation_fps),
            (Actor('glp'), self._recommendation_glp),
            (Actor('gps'), self._recommendation_gps),
            (Actor('kvp'), self._recommendation_kvp),
            (Actor('ldu'), self._recommendation_ldu),
            (Actor('lega'), self._recommendation_lega),
            (Actor('lps'), self._recommendation_lps),
            (Actor('mcg'), self._recommendation_mcg),
            (Actor('pda'), self._recommendation_pda),
            (Actor('poch'), self._recommendation_poch),
            (Actor('rep'), self._recommendation_rep),
            (Actor('sd'), self._recommendation_sd),
            (Actor('sps'), self._recommendation_sps),
            (Actor('svp'), self._recommendation_svp),
            (Actor('csp'), self._recommendation_csp),
        ))

    @cached_property
    def recommendations_associations(self):
        return self.group_recommendations((
            (Actor('eco'), self._recommendation_eco),
            (Actor('sbv-usp'), self._recommendation_sbv_usp),
            (Actor('sgb'), self._recommendation_sgb),
            (Actor('sgv'), self._recommendation_sgv),
            (Actor('travs'), self._recommendation_travs),
            (Actor('vsa'), self._recommendation_vsa),
            (Actor('sav'), self._recommendation_sav),
        ))

    @cached_property
    def has_national_council_share_data(self):
        if self.national_council_election_year:
            return True
        return False

    def file_extract(self, file, locale):
        file = SwissVote.__dict__[file].__get_by_locale__(self, locale)
        return file.extract if file else ''

    @property
    def voting_text_de_CH(self):
        return self.file_extract('voting_text', 'de_CH')

    @property
    def voting_text_fr_CH(self):
        return self.file_extract('voting_text', 'fr_CH')

    @property
    def federal_council_message_de_CH(self):
        return self.file_extract('federal_council_message', 'de_CH')

    @property
    def federal_council_message_fr_CH(self):
        return self.file_extract('federal_council_message', 'fr_CH')

    @property
    def parliamentary_debate_de_CH(self):
        return self.file_extract('parliamentary_debate', 'de_CH')

    @property
    def parliamentary_debate_fr_CH(self):
        return self.file_extract('parliamentary_debate', 'fr_CH')



from onegov.swissvotes import _


class Actor(object):
    """ A helper class to translate political actors. """

    def __init__(self, name):
        self.name = name

    @property
    def abbreviation(self):
        return {
            'acs': _("actor-acs-abbreviation"),
            'auns': _("actor-auns-abbreviation"),
            'bdp': _("actor-bdp-abbreviation"),
            'bpuk': _("actor-bpuk-abbreviation"),
            'csp': _("actor-csp-abbreviation"),
            'cvp-fr': _("actor-cvp-fr-abbreviation"),
            'cvp': _("actor-cvp-abbreviation"),
            'dem': _("actor-dem-abbreviation"),
            'dsp': _("actor-dsp-abbreviation"),
            'eco': _("actor-eco-abbreviation"),
            'edk': _("actor-edk-abbreviation"),
            'edu': _("actor-edu-abbreviation"),
            'endk': _("actor-endk-abbreviation"),
            'evp': _("actor-evp-abbreviation"),
            'fdk': _("actor-fdk-abbreviation"),
            'fdp-fr': _("actor-fdp-fr-abbreviation"),
            'fdp': _("actor-fdp-abbreviation"),
            'fmh': _("actor-fmh-abbreviation"),
            'fps': _("actor-fps-abbreviation"),
            'frap': _("actor-frap-abbreviation"),
            'ga': _("actor-ga-abbreviation"),
            'gdk': _("actor-gdk-abbreviation"),
            'gem': _("actor-gem-abbreviation"),
            'glp': _("actor-glp-abbreviation"),
            'gps': _("actor-gps-abbreviation"),
            'gsoa': _("actor-gsoa-abbreviation"),
            'hev': _("actor-hev-abbreviation"),
            'jcvp': _("actor-jcvp-abbreviation"),
            'jfdp': _("actor-jfdp-abbreviation"),
            'jgps': _("actor-jgps-abbreviation"),
            'jldu': _("actor-jldu-abbreviation"),
            'jlps': _("actor-jlps-abbreviation"),
            'jna': _("actor-jna-abbreviation"),
            'jsvp': _("actor-jsvp-abbreviation"),
            'juso': _("actor-juso-abbreviation"),
            'kdk': _("actor-kdk-abbreviation"),
            'kkjpd': _("actor-kkjpd-abbreviation"),
            'kvp': _("actor-kvp-abbreviation"),
            'ldk': _("actor-ldk-abbreviation"),
            'ldu': _("actor-ldu-abbreviation"),
            'lega': _("actor-lega-abbreviation"),
            'lps': _("actor-lps-abbreviation"),
            'mcg': _("actor-mcg-abbreviation"),
            'pda': _("actor-pda-abbreviation"),
            'poch': _("actor-poch-abbreviation"),
            'pps': _("actor-pps-abbreviation"),
            'psa': _("actor-psa-abbreviation"),
            'rep': _("actor-rep-abbreviation"),
            'sab': _("actor-sab-abbreviation"),
            'sante': _("actor-sante-abbreviation"),
            'sav': _("actor-sav-abbreviation"),
            'sbk': _("actor-sbk-abbreviation"),
            'sblv': _("actor-sblv-abbreviation"),
            'sbv-asb': _("actor-sbv-asb-abbreviation"),
            'sbv-usp': _("actor-sbv-usp-abbreviation"),
            'sd': _("actor-sd-abbreviation"),
            'seks': _("actor-seks-abbreviation"),
            'sgb': _("actor-sgb-abbreviation"),
            'sgv': _("actor-sgv-abbreviation"),
            'skos': _("actor-skos-abbreviation"),
            'sks': _("actor-sks-abbreviation"),
            'smv': _("actor-smv-abbreviation"),
            'sodk': _("actor-sodk-abbreviation"),
            'sol': _("actor-sol-abbreviation"),
            'sps': _("actor-sps-abbreviation"),
            'ssv': _("actor-ssv-abbreviation"),
            'svp': _("actor-svp-abbreviation"),
            'tcs': _("actor-tcs-abbreviation"),
            'travs': _("actor-travs-abbreviation"),
            'vcs': _("actor-vcs-abbreviation"),
            'vdk': _("actor-vdk-abbreviation"),
            'voev': _("actor-voev-abbreviation"),
            'vpod': _("actor-vpod-abbreviation"),
            'vsa': _("actor-vsa-abbreviation"),
        }.get(self.name, self.name.upper())

    @property
    def label(self):
        return {
            'acs': _("actor-acs-label"),
            'auns': _("actor-auns-label"),
            'bdp': _("actor-bdp-label"),
            'bpuk': _("actor-bpuk-label"),
            'csp': _("actor-csp-label"),
            'cvp-fr': _("actor-cvp-fr-label"),
            'cvp': _("actor-cvp-label"),
            'dem': _("actor-dem-label"),
            'dsp': _("actor-dsp-label"),
            'eco': _("actor-eco-label"),
            'edk': _("actor-edk-label"),
            'edu': _("actor-edu-label"),
            'endk': _("actor-endk-label"),
            'evp': _("actor-evp-label"),
            'fdk': _("actor-fdk-label"),
            'fdp-fr': _("actor-fdp-fr-label"),
            'fdp': _("actor-fdp-label"),
            'fmh': _("actor-fmh-label"),
            'fps': _("actor-fps-label"),
            'frap': _("actor-frap-label"),
            'ga': _("actor-ga-label"),
            'gdk': _("actor-gdk-label"),
            'gem': _("actor-gem-label"),
            'glp': _("actor-glp-label"),
            'gps': _("actor-gps-label"),
            'gsoa': _("actor-gsoa-label"),
            'hev': _("actor-hev-label"),
            'jcvp': _("actor-jcvp-label"),
            'jfdp': _("actor-jfdp-label"),
            'jgps': _("actor-jgps-label"),
            'jldu': _("actor-jldu-label"),
            'jlps': _("actor-jlps-label"),
            'jna': _("actor-jna-label"),
            'jsvp': _("actor-jsvp-label"),
            'juso': _("actor-juso-label"),
            'kdk': _("actor-kdk-label"),
            'kkjpd': _("actor-kkjpd-label"),
            'kvp': _("actor-kvp-label"),
            'ldk': _("actor-ldk-label"),
            'ldu': _("actor-ldu-label"),
            'lega': _("actor-lega-label"),
            'lps': _("actor-lps-label"),
            'mcg': _("actor-mcg-label"),
            'pda': _("actor-pda-label"),
            'poch': _("actor-poch-label"),
            'pps': _("actor-pps-label"),
            'psa': _("actor-psa-label"),
            'rep': _("actor-rep-label"),
            'sab': _("actor-sab-label"),
            'sante': _("actor-sante-label"),
            'sav': _("actor-sav-label"),
            'sbk': _("actor-sbk-label"),
            'sblv': _("actor-sblv-label"),
            'sbv-asb': _("actor-sbv-asb-label"),
            'sbv-usp': _("actor-sbv-usp-label"),
            'sd': _("actor-sd-label"),
            'seks': _("actor-seks-label"),
            'sgb': _("actor-sgb-label"),
            'sgv': _("actor-sgv-label"),
            'skos': _("actor-skos-label"),
            'sks': _("actor-sks-label"),
            'smv': _("actor-smv-label"),
            'sodk': _("actor-sodk-label"),
            'sol': _("actor-sol-label"),
            'sps': _("actor-sps-label"),
            'ssv': _("actor-ssv-label"),
            'svp': _("actor-svp-label"),
            'tcs': _("actor-tcs-label"),
            'travs': _("actor-travs-label"),
            'vcs': _("actor-vcs-label"),
            'vdk': _("actor-vdk-label"),
            'voev': _("actor-voev-label"),
            'vpod': _("actor-vpod-label"),
            'vsa': _("actor-vsa-label"),
        }.get(self.name, self.name.upper())

    def html(self, request):
        return '<span title="{}">{}</span>'.format(
            request.translate(self.label),
            request.translate(self.abbreviation)
        )



from cached_property import cached_property
from onegov.swissvotes import _
from onegov.swissvotes.layouts.default import DefaultLayout


class VotesLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Votes")

    @cached_property
    def editbar_links(self):
        if not self.request.has_role('admin', 'editor'):
            return

        result = [
            (
                _("Update dataset"),
                self.request.link(self.model.default(), name='update'),
                'upload-icon'
            ),
            (
                _("Download dataset (CSV)"),
                self.request.link(self.model.default(), name='csv'),
                'export-icon'
            ),
            (
                _("Download dataset (XLSX)"),
                self.request.link(self.model.default(), name='xlsx'),
                'export-icon'
            )
        ]
        if self.request.has_role('admin'):
            result.append((
                _("Delete all votes"),
                self.request.link(self.model.default(), name='delete'),
                'delete-icon'
            ))
        return result

    @cached_property
    def breadcrumbs(self):
        return [
            (_("Homepage"), self.homepage_link, ''),
            (self.title, self.votes_link, 'current'),
        ]


class UpdateVotesLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Update dataset")

    @cached_property
    def breadcrumbs(self):
        return [
            (_("Homepage"), self.homepage_link, ''),
            (_("Votes"), self.votes_link, ''),
            (self.title, '#', 'current'),
        ]


class DeleteVotesLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Delete all votes")

    @cached_property
    def breadcrumbs(self):
        return [
            (_("Homepage"), self.homepage_link, ''),
            (_("Votes"), self.votes_link, ''),
            (self.title, '#', 'current'),
        ]



from babel import Locale
from cached_property import cached_property
from onegov.core.i18n import SiteLocale
from onegov.core.layout import ChameleonLayout
from onegov.swissvotes import _
from onegov.swissvotes.collections import SwissVoteCollection
from onegov.swissvotes.collections import TranslatablePageCollection
from onegov.user import Auth


class DefaultLayout(ChameleonLayout):

    day_long_format = 'skeleton:MMMMd'
    date_long_format = 'long'
    datetime_long_format = 'medium'

    def __init__(self, model, request):
        super().__init__(model, request)
        self.request.include('common')

        self.pages = TranslatablePageCollection(self.request.session)

    @cached_property
    def title(self):
        return ""

    @cached_property
    def top_navigation(self):
        result = [(_("Votes"), self.votes_link)]

        for page in ('dataset', 'about', 'contact'):
            page = self.pages.by_id(page)
            result.append((page.title, self.request.link(page)))

        return result

    @cached_property
    def editbar_links(self):
        return []

    @cached_property
    def breadcrumbs(self):
        return [(_("Homepage"), self.homepage_link, 'current')]

    @cached_property
    def app_version(self):
        return self.app.settings.core.theme.version

    @cached_property
    def static_path(self):
        return self.request.link(self.app.principal, 'static')

    @cached_property
    def homepage_link(self):
        return self.request.link(self.app.principal)

    @cached_property
    def disclaimer_link(self):
        page = self.pages.by_id('disclaimer')
        return page.title, self.request.link(page)

    @cached_property
    def imprint_link(self):
        page = self.pages.by_id('imprint')
        return page.title, self.request.link(page)

    @cached_property
    def votes_link(self):
        return self.request.link(SwissVoteCollection(self.app))

    @cached_property
    def login_link(self):
        if not self.request.is_logged_in:
            return self.request.link(
                Auth.from_request(self.request, to=self.homepage_link),
                name='login'
            )

    @cached_property
    def logout_link(self):
        if self.request.is_logged_in:
            return self.request.link(
                Auth.from_request(self.request, to=self.homepage_link),
                name='logout'
            )

    @cached_property
    def locales(self):
        result = []
        assert self.app.locales == {'de_CH', 'fr_CH', 'en_US'}
        for locale_code in ('de_CH', 'fr_CH', 'en_US'):
            locale = Locale.parse(locale_code)
            result.append((
                locale_code,
                locale.language,
                locale.get_language_name().capitalize(),
                self.request.link(SiteLocale(locale_code, self.request.url))
            ))
        return result

    @cached_property
    def sentry_js(self):
        return self.app.sentry_js



from onegov_testing.utils import create_app
from onegov.core.crypto import hash_password
from onegov.swissvotes import SwissvotesApp
from onegov.swissvotes.models import Principal
from onegov.user import User
from pytest import fixture
from transaction import commit


def pytest_addoption(parser):
    """ Adds a command line argument to scans all the onegov.* sources to
    make sure that the database tables are created.

    Set this option if you run into  sqlalchemy erorrs like:
        "relation XXX does not exist"

    Run it like this:
        py.test src/onegov.swissvotes/onegov/swissvotes/tests/
            -k test_xxx
            --import-scan

    """
    parser.addoption('--import-scan', action="store_true")


def pytest_cmdline_main(config):
    option = config.getoption('--import-scan')
    if option:
        import importscan
        import onegov
        importscan.scan(onegov, ignore=['.test', '.tests'])


def create_swissvotes_app(request, temporary_path):
    app = create_app(
        SwissvotesApp,
        request,
        use_smtp=True,
        depot_backend='depot.io.local.LocalFileStorage',
        depot_storage_path=str(temporary_path),
    )
    app.session_manager.set_locale('de_CH', 'de_CH')

    session = app.session()
    session.add(User(
        username='admin@example.org',
        password_hash=request.getfixturevalue('swissvotes_password'),
        role='admin'
    ))
    session.add(User(
        realname='Publisher',
        username='publisher@example.org',
        password_hash=request.getfixturevalue('swissvotes_password'),
        role='editor'
    ))
    session.add(User(
        realname='Editor',
        username='editor@example.org',
        password_hash=request.getfixturevalue('swissvotes_password'),
        role='member'
    ))

    commit()
    return app


@fixture(scope='session')
def swissvotes_password():
    return hash_password('hunter2')


@fixture(scope="function")
def principal():
    yield Principal()


@fixture(scope="function")
def swissvotes_app(request, temporary_path):
    app = create_swissvotes_app(request, temporary_path)
    yield app
    app.session_manager.dispose()



from onegov.core.security import Private
from onegov.core.security import Public
from onegov.core.templates import render_template
from onegov.swissvotes import _
from onegov.swissvotes import log
from onegov.swissvotes import SwissvotesApp
from onegov.swissvotes.layouts import DefaultLayout
from onegov.swissvotes.layouts import MailLayout
from onegov.user import Auth
from onegov.user import UserCollection
from onegov.user.forms import LoginForm
from onegov.user.forms import PasswordResetForm
from onegov.user.forms import RequestPasswordResetForm
from onegov.user.utils import password_reset_url


@SwissvotesApp.form(
    model=Auth,
    name='login',
    template='login.pt',
    form=LoginForm,
    permission=Public
)
def handle_login(self, request, form):
    """ Handles the login requests. """

    if form.submitted(request):
        response = self.login_to(request=request, **form.login_data)
        form.error_message = _("Wrong username or password")
    else:
        response = None

    return response or {
        'layout': DefaultLayout(self, request),
        'title': _("Login"),
        'form': form,
        'password_reset_link': request.link(
            Auth.from_request(request), name='request-password'
        ),
    }


@SwissvotesApp.html(
    model=Auth,
    name='logout',
    permission=Private
)
def view_logout(self, request):
    """ Handles the logout requests. """

    return self.logout_to(request)


@SwissvotesApp.form(
    model=Auth,
    name='request-password',
    template='form.pt',
    form=RequestPasswordResetForm,
    permission=Public
)
def handle_password_reset_request(self, request, form):
    """ Handles the password reset requests. """

    layout = DefaultLayout(self, request)

    if form.submitted(request):
        users = UserCollection(request.session)
        user = users.by_username(form.email.data)
        if user:
            url = password_reset_url(
                user,
                request,
                request.link(self, name='reset-password')
            )

            request.app.send_transactional_email(
                subject=request.translate(_("Password reset")),
                receivers=(user.username, ),
                reply_to=request.app.mail['transactional']['sender'],
                content=render_template(
                    'mail_password_reset.pt',
                    request,
                    {
                        'title': request.translate(_("Password reset")),
                        'model': None,
                        'url': url,
                        'layout': MailLayout(self, request)
                    }
                )
            )
        else:
            log.info(
                "Failed password reset attempt by {}".format(
                    request.client_addr
                )
            )

        message = _(
            'A password reset link has been sent to ${email}, provided an '
            'account exists for this email address.',
            mapping={'email': form.email.data}
        )
        request.message(message, 'success')
        return request.redirect(layout.login_link)

    return {
        'layout': layout,
        'title': _('Reset password'),
        'form': form,
    }


@SwissvotesApp.form(
    model=Auth,
    name='reset-password',
    template='form.pt',
    form=PasswordResetForm,
    permission=Public
)
def handle_password_reset(self, request, form):
    """ Handles password reset requests. """

    layout = DefaultLayout(self, request)

    if form.submitted(request):
        if form.update_password(request):
            request.message(_("Password changed."), 'success')
            return request.redirect(layout.login_link)
        else:
            form.error_message = _(
                "Wrong username or password reset link not valid any more."
            )
            log.info(
                "Failed password reset attempt by {}".format(
                    request.client_addr
                )
            )

    if 'token' in request.params:
        form.token.data = request.params['token']

    return {
        'layout': layout,
        'title': _('Reset password'),
        'form': form,
    }



from onegov.swissvotes.forms.attachments import AttachmentsForm
from onegov.swissvotes.forms.page import PageForm
from onegov.swissvotes.forms.search import SearchForm
from onegov.swissvotes.forms.update_dataset import UpdateDatasetForm


__all__ = (
    'AttachmentsForm',
    'PageForm',
    'SearchForm',
    'UpdateDatasetForm',
)



from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.expression import FromClause


class values(FromClause):
    """ Productes a Postgres `VALUES` construct.

    Example:
        # FROM (VALUES ((1, 20), (2, 10))) AS lookup (id, count)
        # ORDER BY lookup.count
        query(
            values(
                [column('id', Integer), column('count', Integer)],
                [(1, 20), (2, 10)],
                alias_name='lookup'
            )
        ).order_by('lookup.count')

    Source: https://bitbucket.org/zzzeek/sqlalchemy/wiki/UsageRecipes/PGValues

    """
    named_with_column = True

    def __init__(self, columns, *args, **kw):
        self._column_args = columns
        self.list = args
        self.alias_name = self.name = kw.pop('alias_name', None)

    def _populate_column_collection(self):
        for c in self._column_args:
            c._make_proxy(self)


@compiles(values)
def compile_values(element, compiler, asfrom=False, **kw):
    columns = element.columns
    v = "VALUES {}".format(
        ", ".join(
            "({})".format(
                ", ".join(
                    compiler.render_literal_value(elem, column.type)
                    for elem, column in zip(tup, columns)
                )
            )
            for tup in element.list
        )
    )
    if asfrom:
        if element.alias_name:
            v = "({}) AS {} ({})".format(
                v,
                element.alias_name,
                (", ".join(c.name for c in element.columns))
            )
        else:
            v = f"({v})"
    return v
